#include "types.h"
#include "rom-functions.h"

#define VAR_START ((uint16_t*)0x2d)
#define ARRAY_START ((uint16_t*)0x2f)
#define VAR_END ((uint16_t*)0x31)
#define STRING_START ((uint16_t*)0x64)
#define CURSOR_COLUMN ((uint8_t*)0xd3)

extern void putc(unsigned char c);
extern void integerToFAC(sint16_t n);
extern void printFAC();
extern void floatToFAC(void* adr);

void puts(uint8_t* text)
{
	while (*text) {
		putc(*text);
		text++;
	}
}

typedef struct {
  uint8_t len;
  uint8_t* start;
} string_t;

union variableValue {
	sint16_t integer;
	string_t string;
};

typedef struct {
  uint8_t name1;
  uint8_t name2;
  union variableValue value;
} variable_t;

typedef struct {
  uint8_t name1;
  uint8_t name2;
  uint16_t size;
  uint8_t dimensionCount;
  uint16_t dimension[];
} array_t;

void newline()
{
	putc(0x0d);
	putc(0x0a);
}

void printTypeSuffix(uint8_t type)
{
	switch (type) {
		case 0:
			// float
			break;
		case 1:
			// string
			putc('$');
			break;
		case 3:
			// integer
			putc('%');
			break;
	}				
}

uint16_t swapWord(uint16_t w)
{
	return ((w & 0xff) << 8) | ((w >> 8) & 0xff);
}

void printValue(union variableValue* value, uint8_t type)
{
	uint8_t i;
	switch (type) {
		case 0:
			// float
			floatToFAC(value);
			printFAC();
			break;
		case 1:
			// string
			putc('"');
			for (i = 0; i < value->string.len; i++) {
				putc(value->string.start[i]);
			}
			putc('"');
			break;
		case 3:
			// integer
			integerToFAC(swapWord(value->integer));
			printFAC();
			break;
	}				
}

int main()
{
	uint8_t* varStart = (uint8_t*) *VAR_START;
	uint8_t* arrayStart = (uint8_t*) *ARRAY_START;
	uint8_t* varEnd = (uint8_t*) *VAR_END;
	uint8_t* adr;
	int i;
	for (adr = varStart; adr < arrayStart; adr += 7) {
		variable_t* var = (variable_t*) adr;
		
		// get variable type
		uint8_t type = (var->name1 >> 6) & 2 | (var->name2 >> 7) & 1;

		// skip functions
		if (type == 2) continue;
		
		// print variable name
		putc(var->name1 & 0x7f);
		putc(var->name2 & 0x7f);
		printTypeSuffix(type);

		// align at column 4		
		while (*CURSOR_COLUMN < 4) putc(' ');
		puts("= ");

		// print values
		printValue(&var->value, type);
		newline();
	}
	while (adr < varEnd) {
		uint8_t* arrayValues;
		uint16_t elementCount;
		uint8_t first = 1;
		uint8_t headerSize = 5;
		array_t* array = (array_t*) adr;
		uint8_t typeLength = 0;
		
		// get variable type
		uint8_t type = (array->name1 >> 6) & 2 | (array->name2 >> 7) & 1;
		
		// function arrays are unsupported
		if (type == 2) break;

		// print variable name
		putc(array->name1 & 0x7f);
		putc(array->name2 & 0x7f);
		printTypeSuffix(type);
			
		// get type length
		switch (type) {
			case 0:
				// float
				typeLength = 5;
				break;
			case 1:
				// string
				typeLength = 3;
				break;
			case 3:
				// integer
				typeLength = 2;
				break;
		}

		// print dimensions
		putc('(');
		for (i = array->dimensionCount - 1; i >= 0; i--) {
			uint16_t len = swapWord(array->dimension[i]) - 1;
			if (first) {
				first = !first;
			} else {
				putc(',');
			}
			integerToFAC(len);
			printFAC();
			headerSize += 2;
		}
		puts(") = ");
		
		// print values
		elementCount = (array->size - headerSize) / typeLength;
		arrayValues = adr + headerSize;
		first = 1;
		for (i = 0; i < elementCount; i++) {
			if (first) {
				first = !first;
			} else {
				putc(',');
			}
			printValue((union variableValue*) arrayValues, type);
			arrayValues += typeLength;
		}
		newline();

		adr += array->size;
	}

	newline();
	
	return 0;
}
